
//#ifdef USE_ENCRYPTION

/* Defined to prevent recursive inclusion -------------------------------------*/

//#ifndef _Encrypt_String_H_
//#define _Encrypt_String_H_

#pragma once

#include "Arduino.h"

class EncryptString {

   public:

      EncryptString (void);
      ~EncryptString (void);

      String ShiftEncrypt (String msg, uint8_t val);
      String ShiftDecrypt (String msg, uint8_t val);

      // UTF-8 to Extended ASCII conversion
      void UTF8toExtendedASCII (bool val);
      uint8_t utf8ascii (uint8_t ascii);
      char* utf8ascii (char* s);
      String utf8ascii (String s);

      const char* ISO_DECODER_TYPE = "ISO8859-15";

   private:

      /*
              Point  ISO-8859-1 Removed             ISO-8859-15 Added

              A4  ¤ Currency Sign                   € Euro Sign                              U+20AC
              A6  ¦ Broken Bar                      Š Latin Capital Letter S With Caron      U+0160
              A8  ¨ Diaeresis                       š Latin Small Letter S With Caron        U+0161
              B4  ´ Acute Accent                    Ž Latin Capital Letter Z With Caron      U+017D
              B8  ¸ Cedilla                         ž Latin Small Letter Z With Caron        U+017E
              BC  ¼ Vulgar Fraction One Quarter     Œ Latin Capital Ligature OE              U+0152
              BD  ½ Vulgar Fraction One Half        œ Latin Small Ligature OE                U+0153
              BE  ¾ Vulgar Fraction Three Quarters  Ÿ Latin Capital Letter Y With Diaeresis  U+0178
      */

      String ISO_8859_15[128] = {"", "", "", "", "", "", "", "", "", "",           // 128 to 137 (Not used)
                                 "", "", "", "", "", "", "", "", "", "",           // 138 to 147 (Not used)
                                 "", "", "", "", "", "", "", "", "", "",           // 148 to 157 (Not used)
                                 "", "", " ", "\u00A1", "\u00A2", "\u00A3", "\u20AC", "\u00A5", "\u0160", "\u00A7",   // 158, 159 (Not used)  (160 to 167 used)
                                 "\u0161", "\u00A9", "\u00AA", "\u00AB", "\u00AC", "\u00AD", "\u00AE", "\u00AF", "\u00B0", "\u00B1",  // 168 to 177 (used)
                                 "\u00B2", "\u00B3", "\u017D", "\u00B5", "\u00B6", "\u00B7", "\u017E", "\u00B9", "\u00BA", "\u00BB", // 178 to 187 (used)
                                 "\u0152", "\u0153", "\u0178", "\u00BF", "\u00C0", "\u00C1", "\u00C2", "\u00C3", "\u00C4", "\u00C5", // 188 to 197 (used)
                                 "\u00C6", "\u00C7", "\u00C8", "\u00C9", "\u00CA", "\u00CB", "\u00CC", "\u00CD", "\u00CE", "\u00CF", // 198 to 207 (used)
                                 "\u00D0", "\u00D1", "\u00D2", "\u00D3", "\u00D4", "\u00D5", "\u00D6", "\u00D7", "\u00D8", "\u00D9", // 208 to 217 (used)
                                 "\u00DA", "\u00DB", "\u00DC", "\u00DD", "\u00DE", "\u00DF", "\u00E0", "\u00E1", "\u00E2", "\u00E3", // 218 to 227 (used)
                                 "\u00E4", "\u00E5", "\u00E6", "\u00E7", "\u00E8", "\u00E9", "\u00EA", "\u00EB", "\u00EC", "\u00ED", // 228 to 237 (used)
                                 "\u00EE", "\u00EF", "\u00F0", "\u00F1", "\u00F2", "\u00F3", "\u00F4", "\u00F5", "\u00F6", "\u00F7", // 238 to 247 (used)
                                 "\u00F8", "\u00F9", "\u00FA", "\u00FB", "\u00FC", "\u00FD", "\u00FE", "\u00FF"            // 248 to 255 (used)
                                };

      // This is not a real char, return a character but it can be one or more bytes,
      // so we use String class that is dynamic to collect, we cannot use a single char
      String UTF8_from_ISO8859 (uint8_t tByte) {
         String out = "";
         if (tByte < 128) {
            out += (char) tByte; // Mantain the same ASCII code, just cast to single char
         } else {
            out += ISO_8859_15 [tByte - 128]; // Append to the out string
         }
         return out;
      }



      bool fromUTF8;
};

//#endif  // End of #ifdef USE_ENCRYPTION
