
#include <SPI.h>

//#define USE_SDMMC

#if defined (ESP8266)

   #include <ESP8266WiFi.h>

   #include <SDFS.h>
   #define FileSystem SDFS
   #define SD_CS 0 // 4  //   // <<<<<< Change here SD Chip Select. We use GPIO4 on Wemos D1 Mini   0 on WEMOS D1 MINI SD MINISHIELD, 4 ON WEMOS D32 PRO, ESP32CAM USE SD_MMC

   #warning "----------- NOT A WARNING, JUST A MESSAGE, ESP8266 USE SPI SD WITH CS PIN GPIO0 ------------"

#elif defined (ESP32)

   #include <WiFi.h>

   #include "soc/soc.h"           // To disable brownout detector
   #include "soc/rtc_cntl_reg.h"  // To disable brownout detector

   #ifndef USE_SDMMC
      #include <SD.h>
      #define FileSystem SD
      #define SD_CS 4  // <<<<<< Change here SD Chip Select. Default GPIO4 on Wemos D32 Pro
      #warning "----------- NOT A WARNING, JUST A MESSAGE, ESP32 USE SPI SD WITH CS PIN GPIO4 ------------"
   #else
      #include "FS.h"
      #include "SD_MMC.h"
      #define FileSystem SD_MMC
      #warning "----------- NOT A WARNING, JUST A MESSAGE, ESP32 USE SDMMC ------------"
   #endif

#else

   #error "This sketch is only compatible with ESP8266 and ESP32 modules"

#endif

// Set your network here
#ifndef STASSID
   #define STASSID "PUT-HERE-YOUR-NETWORK-SSID" 
   #define STAPSK  "PUT-HERE-YOUR-NETWORK-PASSWORD" 
#endif

// ---------- SERVER SOCKET ---------

#define MAX_SRV_CLIENTS  3   // How many clients should be able to connect to this ESP

const int ServerPort = 51342;

WiFiServer server (ServerPort);             // Server side
WiFiClient serverClients[MAX_SRV_CLIENTS];  // Client side

bool SocketConnected[MAX_SRV_CLIENTS];

// INFOS
bool SHOW_GENERAL_LOG = true;
bool SHOW_SERIAL_OUT_LOG;
bool SHOW_SERIAL_IN_LOG;
bool SHOW_VERBOSE_LOG;

String mMachineName = "Default";  // Put here your machine name (This is just a default and it is overwritten by the socket)
String mUserName = "User";        // Put here your name         (This is just a default and it is overwritten by the socket)
uint32_t mBaudRate;
uint8_t mCNCType;
uint8_t mFirmware;
uint32_t mNumberOfLines = 0;

bool mDeleteFile = false;  // MESSO A FALSE PER DEBUG, DA IMPOSTARE A TRUE PER ELIMINARE I FILES DOPO LA STAMPA

char mMailServer[128] =   "";     // The SMTP Server              "smtps.tin.it", "mail.tin.it" etc...
char mMailUserName[128] = "";     // The SMTP Server user
char mMailPassword[256] = "";     // The SMTP Server password
char mMailFrom[128] = "";

char mMailTo[128] = "";

bool mMailUseSSL = false;    // If true use SSL socket instead normal insecure
bool mMailStartTLS = false;  // Ignored on this platform

uint16_t mMailServerPort = 0;     // The SMTP Server port          25, 465, 587

#define INFO_DELIMITER "[-#-#-]"

#ifndef FTP_BUFFER_SIZE
   #if defined (ESP8266)
      #define FTP_BUFFER_SIZE  (4096) 
   #elif defined (ESP32)
      #define FTP_BUFFER_SIZE  (8192)  // 16384 TESTED ON ESP32
   #endif

   //   #define FTP_BUFFER_SIZE  (4096)
#endif
static uint8_t buffer[FTP_BUFFER_SIZE] = {0};

// ---------- ENCRYPTION ---------   // NOT USED HERE

#define USE_ENCRYPTION  // JUST ADD SOME SIMPLE ENCRYPTION ON PLAIN TEXT COMMANDS
#ifdef USE_ENCRYPTION
   #include "EncryptString.h"

   EncryptString Encryption;
   // IMPORTANT: Make sure that in the passwords there are no slash characters (\)
   //   #define ENCRYPTION_KEY F("$$Ro#y~iUipr'J=G$>T#<~FQYk83z{o=Q?;o*XiNqgGo)9+lvSgsJ~&;L9@G(0Q<8+:CJqK6@)E74!)~dSMum+:t{wCYKE~M/0U8<Tk]RSU<$-[s?&E/9;F@V<,vuYA8/R{`#~,~yrgi(q#LKS]+j(>ZH&\4gr~y;37|4fQ26MRBNN.:ko3,Lg,:soawfzN.fW}<GXS||$zR^uO9kSf<N|EK/$1R$DBozZG([K5:X=)yWIhVh/zUDhu+!<i-jqh")
   //   #define ENCRYPTION_SEED 1553  // 26
   #define ENCRYPTION_OFFSET 0
#endif

// ------------- SD CARD AND FILES -------

String AppName = "MyApp";

String mDir = "/" + AppName; //+ "/" + FileFolder;
String mFileName = "";
uint32_t mFileSize = 0;

uint32_t TotalDownloadBytes = 0;
uint32_t filePart = 0;

size_t bytesAvail = 0;
size_t bytesIn = 0;
size_t bytesOut = 0;

File writeFile;

bool WaitingInfos = false;
bool FileIsOpenForWrite = false;

uint32_t oldUpProgress = 0;
uint32_t upProgress = 0;

//float TotalSpaceMegaBytes;
float FreeSpaceMegaBytes;

//bool inUpload = false;
bool Working = false;  // Denote the machine is working, eg. printing, machining. On this code always is false

bool WAIT_ACK = true;  // THIS SET THE CONTROL MODE, IF TRUE THE APP WILL WAIT FOR "ok" ACKNOWLEDGE BEFORE IT SEND A NEW LINE, IF FALSE JUST WAIT SOME TIME AND THEN SEND A NEW LINE
//bool ACK_OK = true;
//bool STOP_MOVE = false;
//bool PAUSED = false;

bool success = false;

uint32_t cntSocketOut = 0;

uint32_t StartTime;  // Denote the time when upload start. Used to calculate file transfer speed
uint32_t deltaT;

uint32_t previousMillis;  // Will store last time
uint32_t currentMillis;   // Will store current time

#if defined (ESP8266)
   FSInfo64 info64;
#endif

void setup() {

#if defined (ESP32)
   WRITE_PERI_REG (RTC_CNTL_BROWN_OUT_REG, 0); // Disable brownout detector
#endif

   Serial.begin (115200);
   delay (3000);

   Serial.println ("\n\nInitializing SD Card ...");

   bool success = INITIALIZE_SD();
   if (! success) return;

   Serial.println ("\nFile Path: " + mDir);

   if (! FileSystem.exists (mDir)) {
      Serial.println ("Folder to save files does not exists, create it: " + mDir);

      if (FileSystem.mkdir (mDir)) {
         Serial.println ("Dir created");
      } else {
         Serial.println ("mkdir failed");
      }
   } else {
      Serial.println ("Folder to save files already exists: " + mDir);
      Serial.println ("\nListing files on " + mDir);
      Serial.println (" ");
      // Uncomment to list all files on SD
      //   File root = SD.open ("/");
      File root = SD.open (mDir);
      listFiles (root, 1);
      root.close();
      Serial.println ("\nDone");
   }

   // Connect to AccessPoint
   WiFi.disconnect();
   WiFi.mode (WIFI_STA);
   Serial.print ("\nConnecting to ");
   Serial.println (STASSID);

   // You can remove the password parameter if you AP is open.
   WiFi.begin (STASSID, STAPSK);

   while (WiFi.status() != WL_CONNECTED) {
      delay (500); Serial.print (".");
   }
   Serial.println ("");
   Serial.println ("WiFi connected");
   Serial.print ("Local IP Address: ");
   Serial.println (WiFi.localIP());

   //   delay (1000);
   Serial.println ("Initializing Server on port " + String (ServerPort));
   //   delay (1000);

   //    server.begin (ServerPort);
   server.begin ();
   Serial.println ("Done");

   previousMillis = millis();
}

void loop() {
   currentMillis = millis();

   //   if (FileIsOpenForWrite) {
   ////      if (currentMillis - previousMillis >= 250) {
   ////         Serial.println ("GetSocketNewConnections: " + String (currentMillis - previousMillis));
   ////         previousMillis = currentMillis;
   ////
   ////         GetSocketNewConnections ();
   ////      }
   //   } else {
   //      GetSocketNewConnections ();
   //   }

   if (!FileIsOpenForWrite) {
      GetSocketNewConnections ();
   }

   GetSocketNewData ();

   //   // Check TCP clients for data
   //   if (! FileIsOpenForWrite) {
   //      GetSocketNewData ();
   //   } else {
   //      if (currentMillis - previousMillis >= 10) {  // 10 milliseconds. Do time to transfer a buffer and write to SD. Write is around 180-280 KB/sec
   //         previousMillis = currentMillis;
   //         GetSocketNewData ();
   //      }
   //   }
}

bool INITIALIZE_SD() {

#if defined (ESP8266)
   Serial.println ("\nSD Card ChipSelect pin: " + String (SD_CS));

   SDFSConfig cfg;
   cfg.setAutoFormat (false);
   cfg.setCSPin (SD_CS);
   //cfg.setSPI (SD_SCK_MHZ (50)); // SPI_FULL_SPEED, SD_SCK_MHZ(X) etc ... ORIGINALE
   cfg.setSPI (SD_SCK_MHZ (40)); // SPI_FULL_SPEED, SD_SCK_MHZ(X) etc ...  50 A VOLTE DA PROBLEMI IN SCRITTURA
   FileSystem.setConfig (cfg);

   if (! FileSystem.begin ()) {
      Serial.println ("\nFailed! Is SD Card inserted ? If yes check wires.");
      return false;
   }
   Serial.println ("Done");

#elif defined (ESP32)

#ifndef USE_SDMMC
   Serial.println ("\nSD Card ChipSelect pin: " + String (SD_CS));

   //   if (! FileSystem.begin (SD_CS, SPI, 50000000)) {
   if (! FileSystem.begin (SD_CS, SPI, 40000000)) {
      Serial.println ("\nFailed! Is SD Card inserted ? If yes check wires.");
      return false;
   }
   Serial.println ("Done");
#else
   if (! FileSystem.begin ("/sdcard", true)) {  // SET HERE SD CARD 1BIT MODE (true or false) to free up GPIO12, GPIO13, GPIO4 pins (last is connected to Flash led)
      Serial.println ("\nFailed! Is SD Card inserted ? If yes check wires.");
      return false;
   }
   Serial.println ("Done");
#endif

#endif

#if defined (ESP8266)

#elif defined (ESP32)
   uint8_t cardType = FileSystem.cardType();
   Serial.print ("\nSD Card Type: ");
   switch (cardType) {
      case CARD_MMC:
         Serial.println ("MMC");
         break;
      case CARD_SD:
         Serial.println ("SDSC");
         break;
      case CARD_SDHC:
         Serial.println ("SDHC");
         break;
      default:
         Serial.println ("UNKNOWN");
   }
#endif

   return true;
}

void GetSocketNewConnections () {

   uint8_t i;

   // Find free/disconnected spot
   for (i = 0; i < MAX_SRV_CLIENTS; i++) {

      if (serverClients[i].connected()) {
         //         Serial.println ("\n>>>>>>>> serverClients[" + String (i) + "].connected    SocketConnected[" + String (i) + "] = " + SocketConnected[i]);

         if (! SocketConnected[i] == true) {

            SocketConnected[i] = true;
            Serial.println ("\nClient [" + String (i + 1) + "] Connected");

            // Reset on reconnection
            FileIsOpenForWrite = false;
            WaitingInfos = false;

            // Set socket parameters
            serverClients[i].setNoDelay (true);

            // Set and Get socket parameters
#if defined (ESP8266)
            serverClients[i].setSync (true);
            serverClients[i].setTimeout (120000); // Milliseconds on ESP8266

            bool nodelay = serverClients[i].getNoDelay();
            Serial.printf ("TCP NODELAY: %s\n", nodelay ? "True" : "False");
            bool issync = serverClients[i].getSync();
            Serial.printf ("TCP SYNC: %s\n", issync ? "True" : "False");
            bool keepalive = serverClients[i].isKeepAliveEnabled();
            Serial.printf ("TCP_KEEPALIVE: %s\n", keepalive ? "True" : "False");
            uint16_t keepaliveidle = serverClients[i].getKeepAliveIdle();
            Serial.printf ("TCP KEEPALIVE IDLE: %d\n", keepaliveidle);
            uint16_t keepaliveinterval = serverClients[i].getKeepAliveInterval();
            Serial.printf ("TCP KEEPALIVE INTERVAL: %d\n", keepaliveinterval);
            uint8_t keepalivecount = serverClients[i].getKeepAliveCount();
            Serial.printf ("TCP KEEPALIVE COUNT: %d\n", keepalivecount);
#elif defined (ESP32)
            serverClients[i].setTimeout (120); // Seconds on ESP32

            int flag1 = 0;
            serverClients[i].getOption (0x01, &flag1);
            bool nodelay = flag1;
            Serial.printf ("TCP NODELAY: %s\n", nodelay ? "True" : "False");
            int keepalive = 0;
            serverClients[i].getOption (0x01, &keepalive);
            Serial.printf ("TCP KEEPALIVE: %d\n", keepalive);
#endif

            //            // Hints: server.available() is a WiFiClient with short-term scope when out of scope, a WiFiClient will
            //            // - flush() - all data will be sent
            //            // -  stop() - automatically too
            //            if (i >= MAX_SRV_CLIENTS) {
            //               server.available().println ("Server is busy [Max clients reached]");
            //               Serial.println ("Server is busy with " + String (MAX_SRV_CLIENTS) + " active connections\n");
            //            }

         }
      } else {
         if (SocketConnected[i] == true) {

            SocketConnected[i] = false;
            Serial.println ("\nClient [" + String (i + 1) + "] Disconnected");

            // Reset for next cconnection
            FileIsOpenForWrite = false;
            WaitingInfos = false;
         }
         serverClients[i] = server.available();
      }
   }


}

void GetSocketNewData () {

   //   Serial.println("GetSocketNewData ()");

   bytesAvail = 0;
   bytesIn = 0;
   bytesOut = 0;

   String rcv = "";

   for (int i = 0; i < MAX_SRV_CLIENTS; i++) {

      if (serverClients[i] && serverClients[i].connected()) {

         while ((bytesAvail = serverClients[i].available()) > 0) { // Get received data

            if (FileIsOpenForWrite && (! Working) && (! WaitingInfos)) { // If file is open for write ...

               if (writeFile) {

                  if (bytesAvail > FTP_BUFFER_SIZE) bytesAvail = FTP_BUFFER_SIZE;

                  bytesIn = serverClients[i].read ((uint8_t*)buffer, bytesAvail);
                  //bytesIn = serverClients[i].readBytes ((uint8_t*)buffer, bytesAvail);

                  if (bytesIn > 0) {

                     //                     while ((size_t)writeFile.available() < bytesIn) delay (1);

                     bytesOut = writeFile.write ((uint8_t*)buffer, bytesIn);  // ... write buffer on it
                     if (bytesOut != bytesIn)  Serial.println ("Error while write to file. Read from socket " + String (bytesIn) + " Bytes     Wrote " + String (bytesOut) + " Bytes    " + writeFile.name());

                     TotalDownloadBytes += bytesOut;
                     filePart++;
                  }

                  // GET UPLOAD PROGRESS
                  upProgress = MapLong (TotalDownloadBytes, 0, mFileSize, 0, 100);

                  //      Serial.println ("oldUpProgress: " + String (oldUpProgress) + "    upProgress: " + String (upProgress));

                  if (upProgress != oldUpProgress) {  // PROGRESS PERCENTAGE CHANGED
                     SendSocket ("Progress: " + String (upProgress));

                     if (upProgress == 1) Serial.println (" ");  // Serial log only
                     Serial.println ("File Part: [" + String (filePart) + "]  Progress: " + String (upProgress) + "%  Download: " + String (TotalDownloadBytes) + " - " + String (mFileSize) + "   Elapsed: " + String (deltaT) + " ms, AVG: " + String (getTransferSpeed (TotalDownloadBytes)) + " KBytes/s     Buffer size: " + FTP_BUFFER_SIZE + " Bytes"); // Serial log only

                     //    delay (5);   // <<<<<<<<<<<<<<<<<<<<<<<<< VEDERE SE CI VUOLE

                     oldUpProgress = upProgress;
                  }

                  if (SHOW_VERBOSE_LOG) {
                     Serial.println ("[" + String (millis()) + "]  Collected File Part [" +
                                     String (filePart) + "]  Read from socket: " + String (bytesIn) + " Bytes    Wrote to file: " + String (bytesOut) + " Bytes   Total written: " +
                                     String (TotalDownloadBytes) + " - " + String (mFileSize) + "   Elapsed: " + String (deltaT) + " ms, AVG: " + String (getTransferSpeed (TotalDownloadBytes)) + " KBytes/s");
                  }

                  if (TotalDownloadBytes == mFileSize) {  // Should result equals

                     //                     writeFile.flush();

                     //                     if (writeFile.size() == mFileSize) {
                     //                        GeneralLogMessage ("FILE SIZE TEST1: GOOD: La grandezza dell'upload combacia con la grandezza del file      writeFile.size():  " + String (writeFile.size()) + "     mFileSize: " + String (mFileSize), true, false);
                     //                     } else {
                     //                        GeneralLogMessage ("FILE SIZE TEST1: ERRORE: La grandezza dell'upload non combacia con la grandezza del file     writeFile.size():  " + String (writeFile.size()) + "     mFileSize: " + String (mFileSize), true, false);
                     //                     }

                     Serial.println ("\nClosing file ...");
                     writeFile.close();

                     FileIsOpenForWrite = false;
                     Serial.println ("FileIsOpenForWrite: " + String (FileIsOpenForWrite));

                     Serial.println (" ");
                     Serial.println ("Upload required " + String ((float)deltaT / 1000.0) + " Seconds, AVG: " + String (getTransferSpeed (TotalDownloadBytes)) + " KBytes/s");
                     Serial.println (" ");

                     //                     // DA CONTROLLARE QUUESTI SE GIÀ ESISTONO ALLO START DELL'UPLOAD <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                     //                     TotalDownloadBytes = 0;
                     //                     oldUpProgress = 0;
                     //                     upProgress = 0;

                     String path = mDir + "/" + mFileName;

                     //                     File f = FileSystem.open (path); // ANOTHER TEST
                     //                     if (f.size() == mFileSize) {
                     //                        GeneralLogMessage ("FILE SIZE TEST2: GOOD: La grandezza dell'upload combacia con la grandezza del file      writeFile.size():  " + String (f.size()) + "     mFileSize: " + String (mFileSize), true, false);
                     //                     } else {
                     //                        GeneralLogMessage ("FILE SIZE TEST2: ERRORE: La grandezza dell'upload non combacia con la grandezza del file     writeFile.size():  " + String (f.size()) + "     mFileSize: " + String (mFileSize), true, false);
                     //                     }
                     //                     f.close();

                     if (FileSystem.exists (path)) {  // LAST TEST
                        Serial.println ("\nUPLOAD COMPLETED");  // Serial log omly

                        // Reopen back the file just to test it
#if defined (ESP8266)
                        File f = FileSystem.open (path.c_str(), "r");
#elif defined (ESP32)
                        File f = FileSystem.open (path);
#endif
                        Serial.println ("File SAVED: [" + path + "]   " + String (f.size()) + " Bytes");
                        f.close(); // And close

                        delay (500);
                        SendSocket ("Progress: 0");
                        delay (500);

                        // LIST FILES
                        Serial.println ("\nListing files on " + mDir);
                        Serial.println (" ");
                        File root = SD.open (mDir);
                        listFiles (root, 1);
                        root.close();
                        Serial.println ("\nDone");

                        // LINEA MESSA PROVVISORIA, DA ELIMINARE
                        SendSocket ("COMPLETED: Upload completed in " + String (deltaT) + " ms");

                     } else {

                        Serial.println ("ERROR: Something went wrong while open back saved file (File not exists or file name length equal to zero ?): " + mFileName);

                        // DEBUG
                        Serial.printf ("\nFile name length: %d\n", mFileName.length());
                        Serial.printf ("FileSystem.exists (%s): %s\n", path.c_str(), FileSystem.exists (path) ? "True" : "False");

                        return;
                     }

                  } else if (TotalDownloadBytes > mFileSize) {
                     Serial.println ("ERROR: The file [" + String (writeFile.name()) + "] do not have the right size  Written " + String (TotalDownloadBytes) + " Bytes but the file size must be " + String (mFileSize));

                     Serial.println ("Close file");
                     writeFile.close();

                     FileIsOpenForWrite = false;
                     Serial.println ("FileIsOpenForWrite: " + String (FileIsOpenForWrite));
                     return;
                  }

               } else {
                  Serial.println ("ERROR: The file [" + String (writeFile.name()) + "] is not available for write");
               }


            } else { // End of if (FileIsOpenForWrite)

               rcv = serverClients[i].readStringUntil ('\n');

            }

         }  // End of  while ((bytesAvail = serverClients[i].available()) > 0)

      }  // End of if (serverClients[i] && serverClients[i].connected())

   }  // End of for all clients

   yield();
   //   delay (1);

   if (FileIsOpenForWrite && (! Working) && (! WaitingInfos)) {
      return;
   }

   if (rcv.length() == 0) return;

   //   Serial.printf ("\nParseSocketMessage: [%s]\n", rcv.c_str());
   String s = "";

   // STRING COMMANDS

   //   yield();
   //   ESP.wdtFeed();

   if (rcv.startsWith ("SIZE: ") && (! Working)) {

      //      inUpload = true;

      Serial.println ("\n>>> File SIZE remote command: [" + rcv + "]");  // if (SHOW_VERBOSE_LOG)

      s = "SIZE: ";
      String szStr = rcv.substring (s.length());
      //      mFileSize = szStr.toInt ();

      const char* tmp = szStr.c_str();

      char* pEnd;
      mFileSize = strtoul (tmp, &pEnd, 10);  // Extract file size

      FreeSpaceMegaBytes = GetFreeSpaceMegaBytes(); // Refresh free space left on SD

      //      yield();  // NEED ???

      float mFileSizeMB = 0;
      if (mFileSize > 0) {
         mFileSizeMB = (float)mFileSize / (1024.0 * 1024.0);
         Serial.println ("File size: " + String (mFileSize) + " Bytes = " + String (mFileSizeMB, 2) + " MB of " + FreeSpaceMegaBytes + " MB");
      } else {
         Serial.println ("Error reading remote file size: Zero bytes ???? ");
         return;
      }

      if ((mFileSizeMB + 50) < FreeSpaceMegaBytes)  {  // 50 MB Margin

         SendSocket ("SPACEOK: " + String (FreeSpaceMegaBytes));  // OK THE SPACE LEFT ON SD CARD IS GOOD, FILE CAN BE UPLOADED, INFORM THE SENDER
         WaitingInfos = true;  // Start waiting infos
         Serial.println ("[WAITING INFOS]");  // DEBUG

      } else {

         SendSocket ("NOSPACE: " + String (FreeSpaceMegaBytes));  // NO SPACE LEFT ON SD CARD, INFORM THE SENDER
         //         inUpload = false;
         WaitingInfos = false;  // Stop waiting infos
         return;
      }

   } else if ((! Working) && WaitingInfos) {  // INFO MESSAGE

      // INFO MESSAGE EXAMPLE: "MyFile.gcode-#-#-66638-#-#-115200-#-#-/dev/ttyUSB0-#-#-Creality Ender3-#-#-0-#-#-0-#-#-1-#-#-1-#-#-0-#-#-0-#-#-1-#-#-1-#-#-1-#-#-0-#-#-";

      //      //////////// DEBUG ////////////
      Serial.println ("\nINFO MESSAGE RICEIVED: [" + rcv + "]");
      Serial.println ("LENGHT: " + String (rcv.length()) + " Bytes");

      int oldidx = 0;

      for (int i = 0; i < 21; i++) {  // INFOS CONTAIN 20 ITEMS
         int nextidx = rcv.indexOf (INFO_DELIMITER, oldidx);  // EXTRACT ALL INFOS ARRAY BY SPLITTING THE STRING USING A DELIMITER -#-#-
         if (! nextidx) nextidx = rcv.length();

         //         Serial.printf ("i: %d   oldidx: %d  nextidx: %d\n", i, oldidx, nextidx);

         String tmp = "";

         switch (i) {
            case 0:
               mFileName = rcv.substring (oldidx, nextidx);
               if (SHOW_GENERAL_LOG)  {
                  Serial.println (" ");
                  Serial.println ("INFO: FileName: [" + mFileName + "]");
               }
               break;
            case 1:
               tmp = rcv.substring (oldidx, nextidx);
               mNumberOfLines = tmp.toInt ();
               Serial.println ("INFO: NumOfLines: " + String (mNumberOfLines));
               break;
            case 2:
               tmp = rcv.substring (oldidx, nextidx);
               mBaudRate = tmp.toInt ();
               Serial.println ("INFO: BaudRate: " + String (mBaudRate));
               break;
            case 3: // Force Serial device, not used here
               //    ForcedSerialDevice = rcv.substring (oldidx, nextidx);
               break;
            case 4:
               mMachineName = rcv.substring (oldidx, nextidx);
               Serial.println ("INFO: MachineName: " + mMachineName);
               break;
            case 5:
               tmp = rcv.substring (oldidx, nextidx);
               mCNCType = tmp.toInt ();
               Serial.println ("INFO: CNCType: " + String (mCNCType));
               break;
            case 6:
               tmp = rcv.substring (oldidx, nextidx);
               mFirmware = tmp.toInt ();
               Serial.println ("INFO: Firmware: " + String (mFirmware));
               break;
            case 7:
               //WAIT_ACK = (bool)rcv.substring (oldidx, nextidx);
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? WAIT_ACK = true : WAIT_ACK = false;
               Serial.println ("INFO: WaitACK: " + String (WAIT_ACK));
               break;
            case 8:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? SHOW_GENERAL_LOG = true : SHOW_GENERAL_LOG = false;
               Serial.println ("INFO: GeneralLog: " + String (SHOW_GENERAL_LOG));
               break;
            case 9:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? SHOW_SERIAL_IN_LOG = true : SHOW_SERIAL_IN_LOG = false;
               Serial.println ("INFO: SerialInLog: " + String (SHOW_SERIAL_IN_LOG));
               break;
            case 10:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? SHOW_SERIAL_OUT_LOG = true :  SHOW_SERIAL_OUT_LOG = false;
               Serial.println ("INFO: SerialOutLog: " + String (SHOW_SERIAL_OUT_LOG));
               break;
            case 11:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? SHOW_VERBOSE_LOG = true :  SHOW_VERBOSE_LOG = false;
               Serial.println ("INFO: VerboseLog: " + String (SHOW_VERBOSE_LOG));
               break;
            case 12:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? mDeleteFile = true : mDeleteFile = false;
               Serial.println ("INFO: DeleteFile: " + String (mDeleteFile));
               break;
            case 13:
               mUserName = rcv.substring (oldidx, nextidx);
               if (mUserName.length() == 0) mUserName = "User";
               if (SHOW_GENERAL_LOG)  {
                  Serial.println (" ");
                  Serial.println ("INFO: UserName: [" + mUserName + "]");
               }
               break;


            case 14:
               tmp = rcv.substring (oldidx, nextidx);
               tmp.toCharArray (mMailServer, tmp.length() + 1);
               if (SHOW_GENERAL_LOG)  {
                  Serial.println (" ");
                  Serial.println ("INFO: MailServer: [" + String (mMailServer) + "]");
               }
               break;
            case 15:
               tmp = rcv.substring (oldidx, nextidx);
               //               int p = tmp.toInt ();
               //               if (p > 0) mMailServerPort = p;
               mMailServerPort = tmp.toInt ();
               if (mMailServerPort == 0) mMailServerPort = 587;
               Serial.println ("INFO: MailServerPort: [" + String (mMailServerPort) + "]");
               break;
            case 16:
               tmp = rcv.substring (oldidx, nextidx);
               tmp.toCharArray (mMailUserName, tmp.length() + 1);
               strcpy (mMailFrom, mMailUserName); // Copia username in mail from

               Serial.println ("INFO: MailUserName: [" + String (mMailUserName) + "]");
               break;
            case 17:
               tmp = rcv.substring (oldidx, nextidx);
               tmp.toCharArray (mMailPassword, tmp.length() + 1);
               Serial.println ("INFO: MailPassword: [" + String (mMailPassword) + "]");
               break;
            case 18:
               tmp = rcv.substring (oldidx, nextidx);
               tmp.toCharArray (mMailTo, tmp.length() + 1);
               Serial.println ("INFO: MailTo: [" + String (mMailTo) + "]");
               break;
            case 19:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? mMailUseSSL = true : mMailUseSSL = false;
               Serial.println ("INFO: MailUseSSL: " + String (mMailUseSSL));
               break;
            case 20:
               tmp = rcv.substring (oldidx, nextidx);
               (tmp.toInt () == 1) ? mMailStartTLS = true : mMailStartTLS = false;
               Serial.println ("INFO: MailStartTLS: " + String (mMailStartTLS));
               break;
         }

         //         Serial.printf ("test > [%s]\n", rcv.substring (oldidx).c_str());

         oldidx = nextidx + strlen (INFO_DELIMITER);
      }

      if (! FileIsOpenForWrite) {

         String theFile = mDir + "/" + mFileName;

         Serial.println ("\nCheck if file already exists (if yes delete it): " + theFile);
         if (FileSystem.exists (theFile)) {
            success = FileSystem.remove (theFile); // Remove the file if already exists
            if (success)
               Serial.println ("Deleted old existing file");
            else
               Serial.println ("Cannot delete old existing file");
         } else {
            Serial.println ("The file do not already exists (create it): " + theFile);
         }

         Serial.println ("\nTry to open file in WRITE mode: " + theFile);
         writeFile = FileSystem.open (theFile, "w");  // Open it in append mode
         if (! writeFile) {
            Serial.println ("\nUnable to open file " + theFile + " for write\n");
            return;
         } else {
            Serial.println ("\nFile " + theFile + " is open for write, waiting data ...\n");
         }
      }

      //////// START UPLOAD PROCESS ////////
      Serial.println ("START TIMER");
      StartTime = millis();
      FileIsOpenForWrite = true;
      WaitingInfos = false;
      Serial.println ("FileIsOpenForWrite: " + String (FileIsOpenForWrite));

      // Reset
      TotalDownloadBytes = 0;
      oldUpProgress = 0;
      upProgress = 0;
      filePart = 0;
      deltaT = 0;

      //      return;  // AGGIUNTO DOPO, DA CONTROLLARE

   } else if (rcv.startsWith ("GeneralLog: ")) {
   } else if (rcv.startsWith ("SerialInLog: ")) {
   } else if (rcv.startsWith ("SerialOutLog: ")) {
   } else if (rcv.startsWith ("VerboseLog: ")) {
   } else if (rcv.startsWith ("DeleteFile: ")) {
   } else if (rcv == "REBOOT") {
   } else if (rcv.startsWith ("RELAY_1")) {
   } else if (rcv.startsWith ("RELAY_2")) {
   } else if (rcv == "START") {
   } else if (rcv == "PAUSE") {
   } else if (rcv == "STOP") {
   } else if (rcv.startsWith ("CMD: ")) {

   }

   //   SendSocket ("ok");
}

void SendSocket (String Msg) {

   cntSocketOut++;

   // If SHOW_VERBOSE_LOG Then LogMessage("SOCKET SND: " & Bytes.Length & " Bytes")

   if (Msg.length() == 0 || Msg.length() > 1460) return;

   Msg = Encryption.ShiftEncrypt (Msg, ENCRYPTION_OFFSET);  //  >>>>>>>>> SENZA DELIMITATORE
   Msg += "\n"; // Add a terminator

   for (uint8_t i = 0; i < MAX_SRV_CLIENTS; i++) {  //   if (server.hasClient()) {
      if (SocketConnected[i]) {
         serverClients[i].print (Msg);  // Forse write ?
      }
   }

}

float GetTotalSpaceMegaBytes() {

   //   LogMessage ("[" + String (millis()) + "] GetTotalSpaceMegaBytes", true, false);

   // QUESTA È OK DOPO AVER CORRETTO IL FILE SDFS.h con queste due linee nella funzione info64
   //   ////////////// CORREZIONE BUG CHE TORNA VALORI ERRATI //////////////
   //   //info.totalBytes =_fs.vol()->clusterCount() * info.blockSize;
   //   //info.usedBytes = info.totalBytes - (_fs.vol()->freeClusterCount() * _fs.vol()->sectorsPerCluster() * _fs.vol()->bytesPerSector());
   //   info.totalBytes = (uint64_t) ((uint64_t)_fs.vol()->clusterCount() * (uint64_t)info.blockSize); // MIA MODIFICA, ORA CON I CAST VA BENE
   //   info.usedBytes = (uint64_t) ((uint64_t)info.totalBytes - (uint64_t) ((uint64_t)_fs.vol()->freeClusterCount() * (uint64_t)_fs.vol()->sectorsPerCluster() * (uint64_t)_fs.vol()->bytesPerSector())); // MIA MODIFICA, ORA CON I CAST VA BENE

#if defined (ESP8266)

   //   yield();
   //   ESP.wdtFeed();
   ESP.wdtDisable();

   if (! FileSystem.info64 (info64)) {
      Serial.println ("Error reading SD Card infos");
      //      delay (1000); TimerBlinkError.attach_ms (blinkErrorTime, TimerBlinkError_Tick);
      ESP.wdtEnable (1000);
      return 0;
   }

   ESP.wdtEnable (1000);

   return (float) ((float)info64.totalBytes / (1024.0 * 1024.0));

#elif defined (ESP32)
   yield();
   return (float) ((float)FileSystem.totalBytes() / (1024.0 * 1024.0));
#endif

}

float GetTotalSpaceGigaBytes() {
   return (GetTotalSpaceMegaBytes() / 1024.0);
}

float GetFreeSpaceMegaBytes() {

   uint64_t totalBytes = 0;
   uint64_t usedBytes = 0;

#if defined (ESP8266)

   //   yield();
   //   ESP.wdtFeed();
   ESP.wdtDisable();

   if (! FileSystem.info64 (info64)) {
      Serial.println ("Error reading SD Card infos");
      //      delay (1000); TimerBlinkError.attach_ms (blinkErrorTime, TimerBlinkError_Tick);
      ESP.wdtEnable (1000);
      return 0;
   }

   totalBytes = info64.totalBytes;
   usedBytes = info64.usedBytes;

   ESP.wdtEnable (1000);

#elif defined (ESP32)
   yield();

   uint8_t cardType = FileSystem.cardType();

   if (cardType == CARD_NONE) {
      Serial.println ("No SD card inserted");
      //      delay (1000); TimerBlinkError.attach_ms (blinkErrorTime, TimerBlinkError_Tick);
      return 0;
   }

   totalBytes = FileSystem.totalBytes();
   usedBytes = FileSystem.usedBytes();

#endif

   uint64_t size = totalBytes - usedBytes;
   return (float) ((float)size / (1024.0 * 1024.0));
}

float GetFreeSpaceGigaBytes() {
   return (GetFreeSpaceMegaBytes() / 1024.0);
}

uint16_t getTransferSpeed (uint32_t bytesTransfered) {
   deltaT = (uint32_t) (millis() - StartTime);

   if (bytesTransfered != 0 && deltaT != 0)
      return (bytesTransfered / deltaT);
   else
      return 0;
   //   Serial.println ("Elapsed " + String (deltaT) + " ms,  AVG: " + String (bytesTransfered / deltaT) + " KBytes/s");
}

void listFiles (File dir, int numTabs) {
   while (true) {
      File file = dir.openNextFile();
      if (! file) break;  // no more files

      for (uint8_t i = 0; i < numTabs; i++) Serial.print ('\t');
      Serial.print (file.name());

      if (file.isDirectory()) {
         Serial.println ("/");
         file.close();
         listFiles (file, numTabs + 1);
      } else {  // files have sizes, directories do not
         Serial.print ("\t");
         Serial.print (file.size(), DEC);
         Serial.println (" Bytes");
      }

      file.close();
   }
}


// A BETTER FORM OF ARDUINO map COMMAND

//// Remaps Double number from one range to another.
//double MapDouble (double Value, double fromLow, double fromHigh, double toLow, double toHigh) {
//   return ( (Value - fromLow) * (toHigh - toLow) / (fromHigh - fromLow) + toLow );
//}

// Re-maps a Long number from one range to another.
IRAM_ATTR long MapLong (long Value, long fromLow, long fromHigh, long toLow, long toHigh) {
   //   Serial.printf ("MapLong(%lu, %lu, %lu, %lu, %lu)\n", Value, fromLow, fromHigh, toLow, toHigh);
   return  (Value - fromLow) * (toHigh - toLow) / (fromHigh - fromLow) + toLow;
}
